"""
k3down2 is utility to convert markdown segment into easy to transfer media such as images.
It depends on:

- pandoc to render markdown snippet to html, such as tables.
- graphviz to render graphviz to image.
- playwright (chromium) to render svg/html to png.
- mmdc to convert mermaid chart to svg. See: https://mermaid-js.github.io/mermaid/#
"""

from importlib.metadata import version

__version__ = version("k3down2")
__name__ = "k3down2"

from .down2 import (
    code_to_html,
    convert,
    download,
    graphviz_to_img,
    md_to_html,
    mdtable_to_barehtml,
    mermaid_to_svg,
    render_to_img,
    tex_to_img,
    tex_to_plain,
    tex_to_zhihu,
    tex_to_zhihu_compatible,
    tex_to_zhihu_url,
    web_to_img,
)

__all__ = [
    "code_to_html",
    "convert",
    "download",
    "graphviz_to_img",
    "md_to_html",
    "mdtable_to_barehtml",
    "mermaid_to_svg",
    "render_to_img",
    "tex_to_img",
    "tex_to_plain",
    "tex_to_zhihu",
    "tex_to_zhihu_compatible",
    "tex_to_zhihu_url",
    "web_to_img",
]



#!/usr/bin/env python
# coding: utf-8

from __future__ import annotations

import atexit
import io
import json
import logging
import os
import pathlib
import re
import tempfile
import urllib.error
import urllib.parse
import urllib.request

from PIL import Image, ImageChops
from playwright.sync_api import Browser, sync_playwright
from pylatexenc.latex2text import LatexNodes2Text

import k3proc
from .mime import mime_to_suffix, mimetypes
from .syntax_highlight import code_to_html


logger = logging.getLogger(__name__)

_playwright = None
_browser = None


def _get_browser() -> Browser:
    global _playwright, _browser
    if _browser is None:
        _playwright = sync_playwright().start()
        _browser = _playwright.chromium.launch()
        atexit.register(_shutdown_browser)
    return _browser


def _shutdown_browser() -> None:
    global _playwright, _browser
    if _browser is not None:
        _browser.close()
        _browser = None
    if _playwright is not None:
        _playwright.stop()
        _playwright = None


zhihu_equation_url_fmt = "https://www.zhihu.com/equation?tex={texurl}{align}"

zhihu_equation_fmt = (
    '<img src="https://www.zhihu.com/equation'
    '?tex={texurl}{align}"'
    ' alt="{tex}{altalign}"'
    ' class="ee_img'
    ' tr_noresize"'
    ' eeimg="1">'
)


def convert(input_typ: str, content: str | bytes, output_typ: str, opt: dict[str, dict] | None = None) -> str | bytes:
    conv = mappings.get((input_typ, output_typ))
    if conv is None:
        raise ValueError(f"unsupported conversion: {input_typ} -> {output_typ}")
    if callable(conv):
        kwargs = {}
        if opt is not None and input_typ in opt:
            kwargs = opt[input_typ]
        return conv(content, **kwargs)
    else:
        #  indirect convertion
        inp = convert(input_typ, content, conv, opt=opt)
        return convert(conv, inp, output_typ, opt=opt)


def tex_to_zhihu_compatible(tex: str) -> tuple[str, str]:
    r"""
    Convert tex to zhihu compatible format.
    - ``>`` in img alt mess up the next escaped brace: ``\{ q > 1 \} --> \{ q > 1 }``.
    """

    tex = re.sub(r"\n", "", tex)
    tex = re.sub(r"(?<!\\)>", r"\\gt", tex)
    texurl = urllib.parse.quote(tex)
    return tex, texurl


def _zhihu_tex_params(tex: str, block: bool) -> dict[str, str]:
    """Prepare shared parameters for zhihu equation formatting."""
    tex, texurl = tex_to_zhihu_compatible(tex)
    # zhihu uses double backslash to center-align a block equation.
    align = "%5C%5C" if block else ""
    altalign = "\\\\" if block else ""
    return {"tex": tex, "texurl": texurl, "align": align, "altalign": altalign}


def tex_to_zhihu_url(tex: str, block: bool) -> str:
    """
    Convert tex source to a url linking to a svg on zhihu.
    www.zhihu.com/equation is a public api to render tex into svg.

    Args:
        tex(str): tex source

        block(bool): whether to render a block(center-aligned) equation or
            inline equation.

    Returns:
        string of a ``<img>`` tag.
    """

    p = _zhihu_tex_params(tex, block)
    return zhihu_equation_url_fmt.format(texurl=p["texurl"], align=p["align"])


def tex_to_zhihu(tex: str, block: bool) -> str:
    """
    Convert tex source to a img tag link to a svg on zhihu.
    www.zhihu.com/equation is a public api to render tex into svg.

    Args:
        tex(str): tex source

        block(bool): whether to render a block(center-aligned) equation or
            inline equation.

    Returns:
        string of a ``<img>`` tag.
    """

    p = _zhihu_tex_params(tex, block)
    return zhihu_equation_fmt.format(**p)


superscripts = {
    "0": "⁰",
    "1": "¹",
    "2": "²",
    "3": "³",
    "4": "⁴",
    "5": "⁵",
    "6": "⁶",
    "7": "⁷",
    "8": "⁸",
    "9": "⁹",
    "a": "ᵃ",
    "b": "ᵇ",
    "c": "ᶜ",
    "d": "ᵈ",
    "e": "ᵉ",
    "f": "ᶠ",
    "g": "ᵍ",
    "h": "ʰ",
    "i": "ⁱ",
    "j": "ʲ",
    "k": "ᵏ",
    "l": "ˡ",
    "m": "ᵐ",
    "n": "ⁿ",
    "o": "ᵒ",
    "p": "ᵖ",
    "r": "ʳ",
    "s": "ˢ",
    "t": "ᵗ",
    "u": "ᵘ",
    "v": "ᵛ",
    "w": "ʷ",
    "x": "ˣ",
    "y": "ʸ",
    "z": "ᶻ",
    "A": "ᴬ",
    "B": "ᴮ",
    "D": "ᴰ",
    "E": "ᴱ",
    "G": "ᴳ",
    "H": "ᴴ",
    "I": "ᴵ",
    "J": "ᴶ",
    "K": "ᴷ",
    "L": "ᴸ",
    "M": "ᴹ",
    "N": "ᴺ",
    "O": "ᴼ",
    "P": "ᴾ",
    "R": "ᴿ",
    "T": "ᵀ",
    "U": "ᵁ",
    "V": "ⱽ",
    "W": "ᵂ",
    "+": "⁺",
    "-": "⁻",
    "=": "⁼",
    "(": "⁽",
    ")": "⁾",
}

subscripts = {
    "0": "₀",
    "1": "₁",
    "2": "₂",
    "3": "₃",
    "4": "₄",
    "5": "₅",
    "6": "₆",
    "7": "₇",
    "8": "₈",
    "9": "₉",
    "+": "₊",
    "-": "₋",
    "=": "₌",
    "(": "₍",
    ")": "₎",
    "b": "ᵦ",
    "g": "ᵧ",
    "a": "ₐ",
    "e": "ₑ",
    "i": "ᵢ",
    "j": "ⱼ",
    "o": "ₒ",
    "r": "ᵣ",
    "u": "ᵤ",
    "v": "ᵥ",
    "x": "ₓ",
}


def all_in(chars: str, cate: dict[str, str]) -> bool:
    for c in chars:
        if c not in cate:
            return False
    return True


def tex_to_plain(tex: str) -> str:
    """
    Try hard converting tex to unicode plain text.
    """

    for reg, cate in (
        (r"_\{([^}]*?)\}", subscripts),
        (r"[\^]\{([^}]*?)\}", superscripts),
        (r"_(.)", subscripts),
        (r"[\^](.)", superscripts),
    ):
        pieces = []
        while True:
            match = re.search(reg, tex, flags=re.DOTALL | re.UNICODE)
            if match:
                chars = match.groups()[0]
                if all_in(chars, cate):
                    chars = [cate[x] for x in chars]
                else:
                    chars = tex[match.start() : match.end()]
                pieces.append(tex[: match.start()])
                pieces.append("".join(chars))
                tex = tex[match.end() :]
            else:
                pieces.append(tex)
                break

        tex = "".join(pieces)

    return LatexNodes2Text().latex_to_text(tex)


def tex_to_img(tex: str, block: bool, typ: str) -> str | bytes:
    """
    Convert tex source to an image.

    Args:
        tex(str): tex source

        block(bool): whether to render a block(center-aligned) equation or
            inline equation.

        typ(str): output image type such as "png" or "jpg"

    Returns:
        bytes of png data.
    """

    input_type = "tex_block"
    if not block:
        input_type = "tex_inline"
    return convert(input_type, tex, typ)


def download(url: str) -> bytes:
    """
    Download content from ``url`` and return the responded data.

    Args:
        url(str): the url from which to download.

    Returns:
        bytes of downloaded data.
    """

    resp = urllib.request.urlopen(url, timeout=30)
    return resp.read()


def web_to_img(pagefn: str, typ: str) -> bytes:
    """
    Render a web page, which could be html, svg etc into image.

    Args:
        pagefn(string): path to a local file that can be rendered in a browser.

        typ(string): specify output image type such as "png", "jpg"

    Returns:
        bytes of the image data
    """

    intyp = pagefn.rsplit(".")[-1]
    page = fread(pagefn)
    return render_to_img(intyp, page, typ)


def _trim_and_convert(png_data: bytes, typ: str) -> bytes:
    """Trim whitespace borders from a screenshot and convert to the target format."""
    img = Image.open(io.BytesIO(png_data))

    # Trim borders matching the corner pixel color (like ImageMagick -trim).
    bg = Image.new(img.mode, img.size, img.getpixel((0, 0)))
    diff = ImageChops.difference(img, bg)
    bbox = diff.getbbox()
    if bbox:
        img = img.crop(bbox)

    buf = io.BytesIO()
    if typ == "png":
        img.save(buf, format="PNG")
    else:
        background = Image.new("RGB", img.size, (255, 255, 255))
        if img.mode == "RGBA":
            background.paste(img, mask=img.split()[3])
        else:
            background.paste(img)
        background.save(buf, format="JPEG")

    return buf.getvalue()


def render_to_img(
    mime: str, content: str | bytes, typ: str, width: int = 1000, height: int = 2000, asset_base: str | None = None
) -> bytes:
    """
    Render content that is renderable in a browser to image.
    Such as html, svg etc into image.
    Uses Playwright (Chromium) for rendering and Pillow for image processing.

    Args:
        mime(str): a full mime type such as ``image/jpeg`` or a shortcut ``jpg``.

        content(str): content to render, such as jpeg data or svg source.

        typ(string): specifies output image type such as "png", "jpg"

        width(int): specifies the window width to render a page. Default 1000.

        height(int): specifies the window height to render a page. Default 2000.

        asset_base(str): specifies the path to assets dir. E.g. the image base path in a html page.

    Returns:
        bytes of the image data
    """

    if "html" in mime:
        content = r'<meta http-equiv="Content-Type" content="text/html; charset=utf-8"/>' + content

        if asset_base is not None:
            base_uri = pathlib.Path(asset_base).as_uri()
            content = '<base href="{}/">'.format(base_uri) + content

    m = mimetypes.get(mime) or mime
    suffix = mime_to_suffix.get(m, mime)

    with tempfile.TemporaryDirectory() as tdir:
        fn = os.path.join(tdir, "xxx." + suffix)
        flags = "w"
        if isinstance(content, bytes):
            flags = "wb"
        with open(fn, flags) as f:
            f.write(content)

        browser = _get_browser()
        page = browser.new_page(
            viewport={"width": width, "height": height},
            device_scale_factor=2,
        )
        page.goto(pathlib.Path(fn).as_uri())

        content_height = page.evaluate("document.documentElement.scrollHeight")
        if content_height > height:
            page.set_viewport_size({"width": width, "height": content_height})

        png_data = page.screenshot(omit_background=True)
        page.close()

    return _trim_and_convert(png_data, typ)


html_style = """
<style type="text/css" media="screen">
    table {
        display: block;
        margin-bottom: 1em;
        width: fit-content;
        font-family: -apple-system,BlinkMacSystemFont,"Roboto","Segoe UI","Helvetica Neue","Lucida Grande",Arial,sans-serif;
        font-size: .75em;
        border-collapse: collapse;
        overflow-x: auto;
    }

    thead {
        background-color: #f2f3f3;
        border-bottom: 2px solid #b6b6b6;
    }

    th {
        padding: 0.5em;
        font-weight: bold;
        text-align: left;
    }

    td {
        padding: 0.5em;
        border-bottom: 1px solid #ddd;
    }

    tr,
    td,
    th {
        vertical-align: middle;
    }
    pre.highlight {
        margin: 0;
        padding: 1em;
        background: #263238;
        color: #eff;
        font-size: 1.5em;
        font-family: "SFMono-Regular",Consolas,"Liberation Mono",Menlo,Courier,"PingFang SC", "Microsoft YaHei",monospace;
    }
</style>
"""


def md_to_html(md: str) -> str:
    """
    Build markdown source into html.

    Args:
        md(str): markdown source.

    Returns:
        str of html
    """

    _, html, _ = k3proc.command_ex(
        "pandoc",
        "-f",
        "markdown",
        "-t",
        "html",
        input=md,
    )

    return html_style + html


def mdtable_to_barehtml(md: str) -> str:
    """
    Build markdown table into html without style.

    Args:
        md(str): markdown source.

    Returns:
        str of html
    """

    # A table with wide column will cause pandoc to produce ``colgroup`` tag, which is not recognized by zhihu.
    # Reported in:
    #      https://github.com/drmingdrmer/md2zhihu/issues/22
    #
    # Thus we have to set a very big rendering window to disable this behavior
    #      https://github.com/jgm/pandoc/issues/2574

    _, html, _ = k3proc.command_ex(
        "pandoc",
        "-f",
        "markdown",
        "-t",
        "html",
        "--column",
        "100000",
        input=md,
    )
    lines = html.strip().split("\n")
    lines = [x for x in lines if x not in ("<thead>", "</thead>", "<tbody>", "</tbody>")]

    return "\n".join(lines)


def mermaid_to_svg(mmd: str) -> str:
    """
    Render mermaid to svg.
    See: https://mermaid-js.github.io/mermaid/#

    Requires:
        npm install @mermaid-js/mermaid-cli
    """

    with tempfile.TemporaryDirectory() as tdir:
        output_path = os.path.join(tdir, "mmd.svg")

        puppeteer_config = {
            "args": [
                "--no-sandbox",
                "--disable-setuid-sandbox",
                "--disable-dev-shm-usage",
                "--disable-gpu",
            ]
        }

        config_file_path = os.path.join(tdir, "config.json")
        with open(config_file_path, "w") as f:
            f.write(json.dumps(puppeteer_config))

        k3proc.command_ex(
            "npm",
            "exec",
            "--",
            "mmdc",
            "-o",
            output_path,
            "--puppeteerConfigFile",
            config_file_path,
            input=mmd,
        )
        return fread(output_path)


def graphviz_to_img(gv: str | bytes, typ: str) -> bytes:
    """
    Render graphviz source to image.

    Requires:
        brew install graphviz
    """

    _, out, _ = k3proc.command_ex(
        "dot",
        "-T" + typ,
        input=to_bytes(gv),
        text=False,
    )
    return out


def to_bytes(s: str | bytes) -> bytes:
    if isinstance(s, bytes):
        return s
    return bytes(s, "utf-8")


def pjoin(*p: str) -> str:
    return os.path.join(*p)


def fread(*p: str) -> str:
    with open(os.path.join(*p), "r") as f:
        return f.read()


mappings = {
    ("md", "html"): md_to_html,
    ("md", "jpg"): "html",
    ("md", "png"): "html",
    ("html", "jpg"): lambda x, **kwargs: render_to_img("html", x, "jpg", **kwargs),
    ("html", "png"): lambda x, **kwargs: render_to_img("html", x, "png", **kwargs),
    # markdown table
    ("table", "html"): mdtable_to_barehtml,
    ("table", "jpg"): "html",
    ("table", "png"): "html",
    ("mermaid", "svg"): mermaid_to_svg,
    ("mermaid", "jpg"): "svg",
    ("mermaid", "png"): "svg",
    ("graphviz", "svg"): lambda x: graphviz_to_img(x, "svg"),
    ("graphviz", "jpg"): lambda x: graphviz_to_img(x, "jpg"),
    ("graphviz", "png"): lambda x: graphviz_to_img(x, "png"),
    ("tex_block", "url"): lambda x: tex_to_zhihu_url(x, True),
    ("tex_inline", "url"): lambda x: tex_to_zhihu_url(x, False),
    ("tex_block", "imgtag"): lambda x: tex_to_zhihu(x, True),
    ("tex_inline", "imgtag"): lambda x: tex_to_zhihu(x, False),
    ("url", "jpg"): download,
    ("url", "png"): download,
    ("url", "svg"): download,
    ("url", "html"): download,
    ("tex_block", "jpg"): "svg",
    ("tex_inline", "jpg"): "svg",
    ("tex_block", "png"): "svg",
    ("tex_inline", "png"): "svg",
    ("tex_block", "svg"): "url",
    ("tex_inline", "svg"): "url",
    ("tex_inline", "plain"): tex_to_plain,
    ("svg", "jpg"): lambda x: render_to_img("svg", x, "jpg"),
    ("svg", "png"): lambda x: render_to_img("svg", x, "png"),
    ("code", "html"): code_to_html,
    ("code", "jpg"): "html",
    ("code", "png"): "html",
}



import os
import re
import unittest

import numpy
import k3ut
import skimage
import skimage.io
from skimage.metrics import structural_similarity as ssim

import k3down2

dd = k3ut.dd

this_base = os.path.dirname(__file__)


class TestTex(unittest.TestCase):
    def _clean(self):
        pass

    def setUp(self):
        self._clean()

    def tearDown(self):
        self._clean()

    def test_convert(self):
        d = "test/data/convert"
        for folder, to, opts in (
            (
                "code",
                "jpg",
                {},
            ),
            (
                "code",
                "png",
                {},
            ),
            (
                "code-500",
                "jpg",
                {"width": 500},
            ),
            (
                "code-nolang",
                "jpg",
                {},
            ),
            (
                "code-unicode",
                "jpg",
                {},
            ),
            (
                "md",
                "jpg",
                {},
            ),
            (
                "md",
                "png",
                {},
            ),
            (
                "mermaid",
                "jpg",
                {},
            ),
            (
                "mermaid",
                "png",
                {},
            ),
            (
                "graphviz",
                "jpg",
                {},
            ),
            (
                "graphviz",
                "png",
                {},
            ),
            (
                "table",
                "jpg",
                {},
            ),
            (
                "table",
                "png",
                {},
            ),
            (
                "tex_block",
                "jpg",
                {},
            ),
            (
                "tex_block",
                "png",
                {},
            ),
            (
                "tex_inline",
                "jpg",
                {},
            ),
            (
                "tex_inline",
                "png",
                {},
            ),
        ):
            frm = folder.split("-")[0]
            inp = fread(d, folder, "input")
            got = k3down2.convert(frm, inp, to, opt={"html": opts})

            wantpath = pjoin(d, folder, "want." + to)
            gotpath = pjoin(d, folder, "got." + to)

            fwrite(gotpath, got)

            sim = cmp_image(wantpath, gotpath)
            self.assertGreater(sim, 0.75)

            rm(gotpath)

    def test_tex_to_zhihu_compatible(self):
        tex = r"\{ q > 5, a <> 2 \}"
        want = r"\{ q \gt 5, a <\gt 2 \}"
        wanturl = "%5C%7B%20q%20%5Cgt%205%2C%20a%20%3C%5Cgt%202%20%5C%7D"
        got, goturl = k3down2.tex_to_zhihu_compatible(tex)
        self.assertEqual(want, got)
        self.assertEqual(wanturl, goturl)

    def test_tex_to_zhihu_compatible_preserves_escaped_gt(self):
        cases = [
            # already-escaped \> should not become \\gt
            (r"a \> b", r"a \> b"),
            # arrow -> should not become -\gt
            (r"a -> b", r"a -\gt b"),
            # standalone > still replaced
            (r"x > 1", r"x \gt 1"),
            # multiple unescaped >
            (r"a > b > c", r"a \gt b \gt c"),
        ]
        for tex, want in cases:
            got, _ = k3down2.tex_to_zhihu_compatible(tex)
            self.assertEqual(want, got, tex)

    def test_tex_to_zhihu(self):
        big = r"""
X = \begin{bmatrix}
1      & x_2    & x_2^2 \\

\vdots & \vdots & \vdots \\

1      & x_n    & x_n^2
\end{bmatrix}
""".strip()
        cases = [
            (
                r"a = b",
                True,
                r'<img src="https://www.zhihu.com/equation?tex=a%20%3D%20b%5C%5C" alt="a = b\\" class="ee_img tr_noresize" eeimg="1">',
            ),
            (
                r"a = b",
                False,
                r'<img src="https://www.zhihu.com/equation?tex=a%20%3D%20b" alt="a = b" class="ee_img tr_noresize" eeimg="1">',
            ),
            (
                big,
                True,
                r'<img src="https://www.zhihu.com/equation?tex=X%20%3D%20%5Cbegin%7Bbmatrix%7D1%20%20%20%20%20%20%26%20x_2%20%20%20%20%26%20x_2%5E2%20%5C%5C%5Cvdots%20%26%20%5Cvdots%20%26%20%5Cvdots%20%5C%5C1%20%20%20%20%20%20%26%20x_n%20%20%20%20%26%20x_n%5E2%5Cend%7Bbmatrix%7D%5C%5C" alt="X = \begin{bmatrix}1      & x_2    & x_2^2 \\\vdots & \vdots & \vdots \\1      & x_n    & x_n^2\end{bmatrix}\\" class="ee_img tr_noresize" eeimg="1">',
            ),
            (
                big,
                False,
                r'<img src="https://www.zhihu.com/equation?tex=X%20%3D%20%5Cbegin%7Bbmatrix%7D1%20%20%20%20%20%20%26%20x_2%20%20%20%20%26%20x_2%5E2%20%5C%5C%5Cvdots%20%26%20%5Cvdots%20%26%20%5Cvdots%20%5C%5C1%20%20%20%20%20%20%26%20x_n%20%20%20%20%26%20x_n%5E2%5Cend%7Bbmatrix%7D" alt="X = \begin{bmatrix}1      & x_2    & x_2^2 \\\vdots & \vdots & \vdots \\1      & x_n    & x_n^2\end{bmatrix}" class="ee_img tr_noresize" eeimg="1">',
            ),
        ]

        for tex, block, want in cases:
            got = k3down2.tex_to_zhihu(tex, block)
            self.assertEqual(want, got)

    def test_tex_to_plain(self):
        case = [
            (r"_1", "₁"),
            (r"a_1 + b^2", "a₁ + b²"),
            (r"b^3 a_z pp", "b³ a_z pp"),
            (r"\mathbb{Q}^3", "ℚ³"),
            (r"\mathbb{Q}^{x+1}", "ℚˣ⁺¹"),
            (r"\sum_{1}^{x+1}(i^2)", "∑₁ˣ⁺¹(i²)"),
        ]

        for inp, want in case:
            got = k3down2.tex_to_plain(inp)
            self.assertEqual(want, got, inp)

            got = k3down2.convert("tex_inline", inp, "plain")
            self.assertEqual(want, got, inp)

    def test_tex_to_img(self):
        d = "test/data/tex_to_img_matrix"

        tex = r"""
X = \begin{bmatrix}
1      & x_2    & x_2^2 \\

\vdots & \vdots & \vdots \\

1      & x_n    & x_n^2
\end{bmatrix}
""".strip()

        for typ in ("jpg", "png"):
            wantfn = "want." + typ
            gotfn = "got." + typ
            rm(pjoin(d, gotfn))

            got = k3down2.tex_to_img(tex, False, typ)
            fwrite(d, gotfn, got)
            sim = cmp_image(pjoin(d, wantfn), pjoin(d, gotfn))
            self.assertGreater(sim, 0.75)

            rm(pjoin(d, gotfn))

    def test_tex_to_zhihu_url(self):
        big = r"""
X = \begin{bmatrix}
1      & x_2    & x_2^2 \\

\vdots & \vdots & \vdots \\

1      & x_n    & x_n^2
\end{bmatrix}
""".strip()
        cases = [
            (r"a = b", True, r"https://www.zhihu.com/equation?tex=a%20%3D%20b%5C%5C"),
            (r"a = b", False, r"https://www.zhihu.com/equation?tex=a%20%3D%20b"),
            (
                big,
                True,
                r"https://www.zhihu.com/equation?tex=X%20%3D%20%5Cbegin%7Bbmatrix%7D1%20%20%20%20%20%20%26%20x_2%20%20%20%20%26%20x_2%5E2%20%5C%5C%5Cvdots%20%26%20%5Cvdots%20%26%20%5Cvdots%20%5C%5C1%20%20%20%20%20%20%26%20x_n%20%20%20%20%26%20x_n%5E2%5Cend%7Bbmatrix%7D%5C%5C",
            ),
            (
                big,
                False,
                r"https://www.zhihu.com/equation?tex=X%20%3D%20%5Cbegin%7Bbmatrix%7D1%20%20%20%20%20%20%26%20x_2%20%20%20%20%26%20x_2%5E2%20%5C%5C%5Cvdots%20%26%20%5Cvdots%20%26%20%5Cvdots%20%5C%5C1%20%20%20%20%20%20%26%20x_n%20%20%20%20%26%20x_n%5E2%5Cend%7Bbmatrix%7D",
            ),
        ]

        for tex, block, want in cases:
            got = k3down2.tex_to_zhihu_url(tex, block)
            self.assertEqual(want, got)

    def test_web_to_png(self):
        d = "test/data/svg_to_png_matrix"

        for typ in [
            "png",
            "jpg",
        ]:
            wantfn = "want." + typ
            gotfn = "got." + typ

            rm(d, gotfn)
            data = k3down2.web_to_img(pjoin(d, "input.svg"), typ)
            fwrite(d, gotfn, data)

            sim = cmp_image(os.path.join(d, wantfn), os.path.join(d, gotfn))
            self.assertGreater(sim, 0.75)
            rm(d, gotfn)

    def test_render_to_img(self):
        d = "test/data/render_to_img"

        for frm, to, opts in [
            ("html", "png", {}),
            ("svg", "jpg", {}),
            ("html-code", "png", {}),
            ("html-code-500", "png", {"width": 500}),
            ("html-code-300", "png", {"width": 300}),
        ]:
            frm_typ = frm.split("-")[0]
            typ = to
            gotfn = "got." + typ

            inp = fread(d, frm, "input")

            data = k3down2.render_to_img(frm_typ, inp, to, **opts)
            fwrite(d, frm, gotfn, data)

            sim = cmp_image(os.path.join(d, frm, "want." + typ), os.path.join(d, frm, gotfn))

            self.assertGreater(sim, 0.75)

            rm(d, frm, gotfn)

    def test_download(self):
        url = "https://www.zhihu.com/equation?tex=a%20%3D%20b%5C%5C"

        data = k3down2.download(url)

        with open("test/data/ab.svg", "rb") as f:
            want = f.read()
        self.assertEqual(want, data)

    def test_mdtable_to_barehtml(self):
        md = r"""
| a   | b   | b   |b   |
| :-- | --: | :-: |--- |
| c `foo | bar`   | d   | d   |d   |
| e   | f   | f   |f   |
"""
        got = k3down2.mdtable_to_barehtml(md)
        want = r"""
<table>
<tr class="header">
<th style="text-align: left;">a</th>
<th style="text-align: right;">b</th>
<th style="text-align: center;">b</th>
<th>b</th>
</tr>
<tr class="odd">
<td style="text-align: left;">c <code>foo | bar</code></td>
<td style="text-align: right;">d</td>
<td style="text-align: center;">d</td>
<td>d</td>
</tr>
<tr class="even">
<td style="text-align: left;">e</td>
<td style="text-align: right;">f</td>
<td style="text-align: center;">f</td>
<td>f</td>
</tr>
</table>
""".strip()

        want = normalize_pandoc_output(want, got)

        self.assertEqual(want, got)

    def test_mdtable_to_barehtml_wild_table(self):
        """
        A table with wide column will cause pandoc to produce ``colgroup`` tag, which is not recognized by zhihu.
        Reported in:
        https://github.com/drmingdrmer/md2zhihu/issues/22

        Thus we have to set a very big rendering window to disable this behavior
            https://github.com/jgm/pandoc/issues/2574

        """

        md = r"""
| a   | b   |
| :-- | --: |
| yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy     | d   |
| e   | f   |
"""
        got = k3down2.mdtable_to_barehtml(md)
        want = r"""
<table>
<tr class="header">
<th style="text-align: left;">a</th>
<th style="text-align: right;">b</th>
</tr>
<tr class="odd">
<td style="text-align: left;">yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy yyyy</td>
<td style="text-align: right;">d</td>
</tr>
<tr class="even">
<td style="text-align: left;">e</td>
<td style="text-align: right;">f</td>
</tr>
</table>
""".strip()

        want = normalize_pandoc_output(want, got)

        self.assertEqual(want, got)

    def test_md_to_html(self):
        md = r"""
| a   | b   | b   |b   |
| :-- | --: | :-: |--- |
| c `foo | bar`   | d   | d   |d   |
| e   | f   | f   |f   |
"""
        got = k3down2.md_to_html(md)
        want = r"""
<style type="text/css" media="screen">
    table {
        display: block;
        margin-bottom: 1em;
        width: fit-content;
        font-family: -apple-system,BlinkMacSystemFont,"Roboto","Segoe UI","Helvetica Neue","Lucida Grande",Arial,sans-serif;
        font-size: .75em;
        border-collapse: collapse;
        overflow-x: auto;
    }

    thead {
        background-color: #f2f3f3;
        border-bottom: 2px solid #b6b6b6;
    }

    th {
        padding: 0.5em;
        font-weight: bold;
        text-align: left;
    }

    td {
        padding: 0.5em;
        border-bottom: 1px solid #ddd;
    }

    tr,
    td,
    th {
        vertical-align: middle;
    }
    pre.highlight {
        margin: 0;
        padding: 1em;
        background: #263238;
        color: #eff;
        font-size: 1.5em;
        font-family: "SFMono-Regular",Consolas,"Liberation Mono",Menlo,Courier,"PingFang SC", "Microsoft YaHei",monospace;
    }
</style>
<table>
<thead>
<tr class="header">
<th style="text-align: left;">a</th>
<th style="text-align: right;">b</th>
<th style="text-align: center;">b</th>
<th>b</th>
</tr>
</thead>
<tbody>
<tr class="odd">
<td style="text-align: left;">c <code>foo | bar</code></td>
<td style="text-align: right;">d</td>
<td style="text-align: center;">d</td>
<td>d</td>
</tr>
<tr class="even">
<td style="text-align: left;">e</td>
<td style="text-align: right;">f</td>
<td style="text-align: center;">f</td>
<td>f</td>
</tr>
</tbody>
</table>
"""
        want = normalize_pandoc_output(want, got)

        self.assertEqual(want, got)

    def test_md_to_html_embed_image(self):
        # Render a markdown of a single table with a image in it.
        # By specifying the asset base url, it should be rendered correctly.

        md = r"""
| a   | b   | b   |b   |
| :-- | --: | :-: |--- |
| ![](foo.jpg)   | d   | d   |d   |
| e   | f   | f   |f   |
"""
        d = "test/data/md_to_html"
        gothtml = k3down2.md_to_html(md)

        data = k3down2.convert(
            "html",
            gothtml,
            "jpg",
            opt={"html": {"asset_base": os.path.abspath(d) + "/assets"}},
        )
        fwrite(d, "got.jpg", data)

        sim = cmp_image(os.path.join(d, "want.jpg"), os.path.join(d, "got.jpg"))
        self.assertGreater(sim, 0.75)

        rm(d, "got.jpg")


def normalize_pandoc_output(want, got):
    #  pandoc may output different style html:
    #  '<tab[25 chars]n<th style="text-align: left;">a</th>\n<th sty[427 chars]ble>' !=
    #  '<tab[25 chars]n<th align="left">a</th>\n<th align="right">b<[310 chars]ble>'

    if 'align="' in got:
        want = re.sub(r'style="text-align: (left|right|center);"', r'align="\1"', want)
    return want


def cmp_image(want, got):
    from PIL import Image as PILImage

    want_img = PILImage.open(want).convert("RGB")
    got_img = PILImage.open(got).convert("RGB")

    # Resize got to match want dimensions if they differ
    if want_img.size != got_img.size:
        got_img = got_img.resize(want_img.size, PILImage.LANCZOS)

    da = numpy.array(want_img)
    db = numpy.array(got_img)

    img1 = skimage.img_as_int(da)
    img2 = skimage.img_as_int(db)

    print("img1:-------------", want)
    print(img1.shape)
    print("img2:-------------", got)
    print(img2.shape)

    p = ssim(img1, img2, channel_axis=2)

    print("similarity(want/got):", want, got, p)
    return p


def pjoin(*p):
    return os.path.join(*p)


def rm(*p):
    try:
        os.unlink(os.path.join(*p))
    except OSError:
        pass


def fread(*p):
    with open(os.path.join(*p), "r") as f:
        return f.read()


def fwrite(*p):
    cont = p[-1]
    p = p[:-1]
    with open(os.path.join(*p), "wb") as f:
        f.write(cont)


def is_ci():
    # github ci
    return os.environ.get("CI") is not None



from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter
from pygments.style import Style
from pygments.token import (
    Keyword,
    Name,
    Comment,
    String,
    Error,
    Text,
    Number,
    Operator,
    Punctuation,
    Literal,
)


base00 = "#263238"
base01 = "#2e3c43"
base02 = "#314549"
base03 = "#546e7a"
base04 = "#b2ccd6"
base05 = "#eeffff"
base06 = "#eeffff"
base07 = "#ffffff"
base08 = "#f07178"
base09 = "#f78c6c"
base0a = "#ffcb6b"
base0b = "#c3e88d"
base0c = "#89ddff"
base0d = "#82aaff"
base0e = "#c792ea"
base0f = "#ff5370"


class Base16Style(Style):
    background_color = base00
    highlight_color = base02
    default_style = base05

    styles = {
        Text: base05,
        Error: base08,
        Comment: base03,
        Keyword: base0e,
        Keyword.Type: base0a,
        Keyword.Constant: base0d,
        Keyword.Namespace: base0d,
        Name: base05,
        Name.Builtin: base0a,
        Name.Function: base0d,
        Name.Class: base0d,
        Name.Decorator: base0e,
        Name.Exception: base08,
        Number: base09,
        Operator: base0d,
        Operator.Word: base0d,
        Punctuation: base05,
        Literal: base0b,
        String: base0b,
    }


def escape(s, quote=True):
    s = s.replace("&", "&amp;")
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    if quote:
        s = s.replace('"', "&quot;")
    return s


def code_to_html(text):
    lines = text.strip().split("\n")

    # extract code language: ```go
    lang = lines[0][3:].strip()
    text = "\n".join(lines[1:-1])

    linenos = False
    style = Base16Style

    lineheight = "1.8"

    if lang == "":
        # For illustration text graph, use lower line height.
        # Becuase in such case it prefers an easy to read digram than a
        # comforable line height.
        lineheight = "1.3"

    prestyles = (
        r"line-height: {} !important;"
        " margin: 0 !important;"
        " padding: 1em;"
        " white-space: pre-wrap;"
        " background: {};"
        " color: {};"
    ).format(
        lineheight,
        style.background_color,
        style.default_style,
    )

    if lang == "":
        text = text.strip()
        return '<pre style="%s"><code>%s</code></pre>\n' % (prestyles, escape(text))

    try:
        lexer = get_lexer_by_name(lang, stripall=True)
        formatter = HtmlFormatter(noclasses=True, linenos=linenos, style=style, prestyles=prestyles)
        code = highlight(text, lexer, formatter)
        if linenos:
            return '<div class="highlight-wrapper">%s</div>\n' % code
        return code
    except Exception:
        return '<pre class="%s"><code>%s</code></pre>\n' % (lang, escape(text))
